
import logging

import requests
import six

import stscraper as scraper
import stutils
import stutils.decorators as d
from stutils import versions

TIMEOUT = stutils.get_config('PYPI_TIMEOUT', 10)
urlretrieve = six.moves.urllib.request.urlretrieve
logger = logging.getLogger('stecosystems')


class PackageDoesNotExist(ValueError):
    pass


class BasePackage(object):
    base_url = None
    name = None  # package name

    @classmethod
    def all(cls, **kwargs):
        # type: (**dict) -> BasePackage
        raise NotImplementedError

    def __init__(self, name, **kwargs):
        """
        Args:
            name (str): package name.
                Expect it will be normalized to prevent duplicate packages.
            info (Optional[dict]): in some cases (I'm talking about you, npm)
                ecosystem API lists packages with their metadata, so it's
                cheaper to reuse it.
                End users should not use this parameter.
        """
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<?? package: %s>" % self.name

    @classmethod
    def _request(cls, *path):
        if path and not (path[0].startswith('https://') or path[0].startswith('http://')):
            path = (cls.base_url,) + path

        for _ in range(3):
            try:
                r = requests.get("/".join(path), timeout=TIMEOUT)
            except requests.exceptions.Timeout:
                continue
            r.raise_for_status()
            return r
        raise IOError("Failed to reach %s. "
                      "Check your Internet connection." % cls.base_url)

    def releases(self, include_unstable=False, include_backports=False):
        """ Return package release labels

        Args:
            include_unstable (bool): whether to include releases including
                symbols other than dots and numbers.
            include_backports (bool): whether to include releases smaller in
                version than last stable release

        Returns:
             List[Tuple[str, str]]: (label, date), sorted by date
        """
        raise NotImplementedError

    def download_url(self, ver):
        """Get URL to package file of the specified version
        This function takes into account supported file types and their
        relative preference (e.g. wheel files before source packages)

        Args:
            ver (str): version string

        Returns:
            Optional[str]: url string if found, None otherwise
        """
        raise NotImplementedError

    def download(self, ver=None):
        """Download and extract the specified package version

        Args:
             ver (str): Version of the package

        Returns:
            Optional[str]: path to the folder with extracted package,
                None if download failed
        """
        raise NotImplementedError

    @property
    def repository(self):
        """ Search for software repository URL

        Returns:
            Optional[str]: path to the folder with extracted package,
                None if download failed
        """
        raise NotImplementedError

    def dependencies(self, ver=None):
        """ Get technical dependencies

        Args:
            ver (Optional[str]): version string, latest version by default

        Returns:
            Dict[str, str]: dictionary of the form `{package: version}`;
                `version` might contain platform-specific modifiers, e.g.
                '>=3.12.1'. If not specified, `version will be `None`.
        """
        raise NotImplementedError

    def loc_size(self, ver):
        """ Get package size in LOC """
        raise NotImplementedError


def resolve_field(item, field, default=None):
    """Retrieve a field from JSON structure

    >>> item = {'one': 1, 'two': [1], 'three': {}}
    >>> resolve_field(item, 'one')
    1
    >>> resolve_field(item, 'two')
    1
    >>> resolve_field(item, 'three')
    {}
    >>> resolve_field(item, 'four') is None
    True
    """
    if not item:
        return default
    if isinstance(item, list):
        res = resolve_field(item[0], field)
    elif isinstance(item, six.string_types):
        res = item
    else:
        res = item.get(field, default)

    if isinstance(res, list):
        return res[0]
    return res


def json_path(item, *path):
    # type: (dict, *str) -> object
    """Helper function to traverse JSON

    >>> a = {'doc': {'versions': {'0.1': {'time': '2018-01-01T00:00.00.00Z'}}}}
    >>> json_path(a, 'doc', 'versions', '0.1', 'time')
    '2018-01-01T00:00.00.00Z'
    >>> json_path(a, 'doc', 'times', '0.1', 'time') is None
    True
    """
    res = item
    for key in path:
        try:  # this way is faster and supports list indexes
            res = res[key]
        except (IndexError, KeyError):
            return None
    return res



#!/usr/bin/env python

"""
API docs:
    https://github.com/npm/registry/blob/master/docs/REGISTRY-API.md
API base URL:
    http://registry.npmjs.com/

extra metrics:
    https://api-docs.npms.io/#api-Package-GetMultiPackageInfo

"""

from __future__ import print_function

import urllib

import ijson.backends.yajl2 as ijson

from .base import *
from stutils import decorators as d

fs_cache = d.fs_cache('npm')


class Package(BasePackage):
    base_url = 'http://registry.npmjs.com/'
    info = None  # stores cached package info

    @classmethod
    def all(cls, cache_file=None):
        if isinstance(cache_file, six.string_types):
            fh = open(cache_file, 'rb')
        elif isinstance(cache_file, six.StringIO):
            fh = cache_file
        else:
            # how to create cache file: wget -O npm.json <url below>
            # it is 14Gb as of Jan 2019
            fh = urllib.urlopen(
                'https://skimdb.npmjs.com/registry/_all_docs?include_docs=true')

        for package_info in ijson.items(fh, 'rows.item'):
            package_name = package_info['id']
            yield Package(package_name, info=package_info['doc'])

    def __init__(self, name, info=None):
        """
        Args:
            name (str): package name.
                Expect it will be normalized to prevent duplicate packages.
            info (Optional[dict]): in some cases (I'm talking about you, npm)
                ecosystem API lists packages with their metadata, so it's
                cheaper to reuse it.
                End users should not use this parameter.
        """
        if info:
            self.info = info
        else:
            try:
                self.info = self._request(name).json()
            except IOError:
                raise PackageDoesNotExist(
                    "Package %s does not exist in npm" % name)

        super(Package, self).__init__(name)

    @d.cached_property
    def _extra_info(self):
        return requests.get(
            "https://api.npms.io/v2/package/" + self.name).json()

    @property
    def quality(self):
        return json_path(self._extra_info, 'score', 'detail', 'quality')

    @property
    def popularity(self):
        return json_path(self._extra_info, 'score', 'detail', 'popularity')

    @property
    def maintenance_score(self):
        return json_path(self._extra_info, 'score', 'detail', 'maintenance')

    @classmethod
    def _request(cls, *path):
        if path and not (path[0].startswith('https://') or path[0].startswith('http://')):
            path = (cls.base_url,) + path

        for _ in range(3):
            try:
                r = requests.get("/".join(path), timeout=TIMEOUT)
            except requests.exceptions.Timeout:
                continue
            r.raise_for_status()
            return r
        raise IOError("Failed to reach %s. "
                      "Check your Internet connection." % cls.base_url)

    def releases(self, include_unstable=False, include_backports=False):
        """ Return package release labels

        Args:
            include_unstable (bool): whether to include releases including
                symbols other than dots and numbers.
            include_backports (bool): whether to include releases smaller in
                version than last stable release

        Returns:
             List[Tuple[str, str]]: (label, date), sorted by date
        """
        raise NotImplementedError

    def download_url(self, ver):
        """Get URL to package file of the specified version
        This function takes into account supported file types and their
        relative preference (e.g. wheel files before source packages)

        Args:
            ver (str): version string

        Returns:
            Optional[str]: url string if found, None otherwise
        """
        assert ver in self.info['releases']
        raise NotImplementedError

    def download(self, ver=None):
        """Download and extract the specified package version

        Args:
             ver (str): Version of the package

        Returns:
            Optional[str]: path to the folder with extracted package,
                None if download failed
        """
        raise NotImplementedError

    @d.cached_property
    def repository(self):
        """ Search for software repository URL

        Returns:
            Optional[str]: path to the folder with extracted package,
                None if download failed
        """
        raise NotImplementedError

    def dependencies(self, ver=None):
        """ Get technical dependencies

        Args:
            ver (Optional[str]): version string, latest version by default

        Returns:
            Dict[str, str]: dictionary of the form `{package: version}`;
                `version` might contain platform-specific modifiers, e.g.
                '>=3.12.1'. If not specified, `version will be `None`.
        """
        raise NotImplementedError

    def loc_size(self, ver):
        """ Get package size in LOC """
        raise NotImplementedError



#!/usr/bin/env python

""" An abstraction of Python package repository API (PyPi API).

"""

from __future__ import print_function

import json
import os
import re
import shutil
from xml.etree import ElementTree

from .base import *
from stutils import sysutils

DEFAULT_SAVE_PATH = '/tmp/pypi'
# directory where package archives are stored
PYPI_SAVE_PATH = stutils.get_config('PYPI_SAVE_PATH', DEFAULT_SAVE_PATH)
sysutils.mkdir(PYPI_SAVE_PATH)

logger = logging.getLogger("ghd.pypi")
fs_cache = d.fs_cache('pypi')

# path to provided shell scripts
_PATH = os.path.dirname(__file__) or '.'


def shell(cmd, *args, **kwargs):
    if kwargs.get('local', True):
        del kwargs['local']
        kwargs['rel_path'] = _PATH
    return sysutils.shell(cmd, *args, **kwargs)


# supported formats and extraction commands
UNZIP = 'unzip -qq -o "%(fname)s" -d "%(dir)s" 2>/dev/null'
UNTGZ = 'tar -C "%(dir)s" --strip-components 1 -zxf "%(fname)s" 2>/dev/null'
UNTBZ = 'tar -C "%(dir)s" --strip-components 1 -jxf "%(fname)s" 2>/dev/null'

SUPPORTED_FORMATS = {
    '.zip': UNZIP,
    '.whl': UNZIP,
    '.egg': UNZIP,  # can't find a single package to test. Are .eggs extinct?
    '.tar.gz': UNTGZ,
    '.tgz': UNTGZ,
    '.tar.bz2': UNTBZ,
    # rpm: contain directory structure from the root and thus can't be parsed
    # (same issue with bdist_dumb packages)
    # '.rpm': 'rpm2cpio "%(fname)s" | $(cd "%(dir)s" && cpio -i -d)'
}

"""
Notes:
1. There is no reliable source for supported Python version.
    requires-dist format is described here:
        https://www.python.org/dev/peps/pep-0345/#version-specifiers
    Unfortunately, it is not informative at all:
        - vast majority (99%) of packages doesn't use it
        - many of those which use do not conform to the standard
"""


class PackageDoesNotExist(ValueError):
    pass


def get_builtins(python_version):
    """ Return set of built-in libraries for Python2/3 respectively
    Intented for parsing imports from source files
    """
    assert python_version in (2, 3)
    url = "https://docs.python.org/%s/library/index.html" % python_version
    text = requests.get(url, timeout=TIMEOUT, verify=False).text
    # text is html and can't be processed with Etree, so regexp it is
    return set(b for b in re.findall(
        r"""<span\s+class=["']pre["']\s*>\s*([\w_-]+)\s*</span>""", text))


def python_loc_size(package_dir):
    """ Get LOC size of a given project

    The LOC count is parsed from pylint output, which includes this table:

    Raw metrics
    -----------

    +----------+-------+------+---------+-----------+
    |type      |number |%     |previous |difference |
    +==========+=======+======+=========+===========+
    |code      |4882   |61.13 |190351   |-185469.00 |
    +----------+-------+------+---------+-----------+
    |docstring |1134   |14.20 |37023    |-35889.00  |
    +----------+-------+------+---------+-----------+
    |comment   |282    |3.53  |17465    |-17183.00  |
    +----------+-------+------+---------+-----------+
    |empty     |1688   |21.14 |52558    |-50870.00  |
    +----------+-------+------+---------+-----------+

    In this example (pandas 0.1) we're looking for 4882
    """
    status, pylint_out = shell("pylint", "--py3k", package_dir,
                               local=False, raise_on_status=False,
                               stderr=open(os.devnull, 'w'))
    if status == 2:
        raise EnvironmentError("pylint is not installed (just in case, path is "
                               "%s)" % package_dir)
    m = re.search(r"\|code\s*\|([\s\d]+?)\|", pylint_out)
    match = m and m.group(1).strip()
    if not match:
        return 0
    return int(match)


class Package(BasePackage):
    base_url = "https://pypi.org"
    info = None  # stores cached package info
    _dirs = None  # created directories to cleanup later

    @classmethod
    def all(cls):
        tree = ElementTree.fromstring(cls._request("simple/").content)
        for package_name in sorted(a.text.lower() for a in tree.iter('a')):
            try:
                package = Package(package_name)
            except PackageDoesNotExist:
                continue
            else:
                yield package

    def __init__(self, name, **kwargs):
        try:
            self.info = self._request("pypi", self.name, "json").json()
        except IOError:
            raise PackageDoesNotExist(
                "Package %s does not exist on PyPi" % name)
        except ValueError:  # simplejson.scanner.JSONDecodeError is a subclass
            # malformed json
            raise ValueError("PyPi package description is invalid")

        self._dirs = []
        self.latest_ver = self.info['info'].get('version')
        super(Package, self).__init__(self.info['info']['name'])

    def __del__(self):
        if DEFAULT_SAVE_PATH != PYPI_SAVE_PATH:
            return
        for folder in self._dirs:
            try:
                # str conversion is required because of this shutil bug:
                # https://bugs.python.org/issue24672
                # use tai5_uan5_gian5_gi2_tsu1_liau7_khoo3-tng7_su5_piau1_im1
                # to test this issue
                shutil.rmtree(str(folder))
            except OSError:
                logger.debug("Error removing temp dir after package %s: %s",
                             self.name, folder)

    def __str__(self):
        return self.canonical_name

    def __repr__(self):
        return "<PyPi package: %s>" % self.name

    def releases(self, include_unstable=False, include_backports=False):
        """Return release labels
        :param include_unstable: bool, whether to include releases including
            symbols other than dots and numbers
        :param include_backports: bool, whether to include releases smaller in
            version than last stable release
        :return list of (label, date), sorted by date

        >>> len(Package("django").releases()) > 10
        True
        >>> len(Package("django").releases()[0])
        2
        >>> isinstance(Package("django").releases()[0], tuple)
        True
        """
        releases = sorted([
            (label, min(f['upload_time'][:10] for f in files))
            for label, files in self.info['releases'].items()
            if files],  # skip empty releases
            key=lambda r: r[1])  # sort by date

        if not include_unstable:
            releases = [(label, date)
                        for label, date in releases
                        if re.match(r"^\d+(\.\d+)*$", label)]

        if not include_backports and releases:
            _rel = []
            for label, date in releases:
                if not _rel or versions.compare(label, _rel[-1][0]) >= 0:
                    _rel.append((label, date))
            releases = _rel

        return releases

    def download_url(self, ver):
        """Get URL to package file of the specified version
        This function takes into account supported file types and their
        relative preference (e.g. wheel files before source packages)

        :param ver: str, version string
        :return: url string if found, None otherwise
        """
        assert ver in self.info['releases']
        # the rationale for iterating several times filtering out pkgtype:
        # some formats are more expensive to process, so it is basically
        # a preference order
        # NOT SUPPORTED: "bdist_dumb", "bdist_rpm", "bdist_deb", "bdist_wininst"
        #   bdist_dumb (contains file structure from root)
        #   bdist_rpm - need to add cpio command to extract content
        #   bdist_deb - not enough data / haven't seen any so far
        #   bdist_wininst is .exe files, most of the time
        # The last three remain on the list only because they
        # often contain source dist instead
        for pkgtype in ("bdist_wheel", "bdist_egg", "sdist",
                        "bdist_rpm", "bdist_deb", "bdist_wininst"):
            for info in self.info['releases'][ver]:
                if info['packagetype'] == pkgtype and \
                    any(info['url'].endswith(ext)
                        for ext in SUPPORTED_FORMATS):
                    return info['url']
        # no downloadable files in supported format
        logger.info("No downloadable files in supported formats "
                    "for package %s ver %s found", self.name, ver)
        return None

    @d.cached_method
    def download(self, ver=None):
        """Download and extract the specified package version from PyPi
        :param ver - Version of package
        """
        ver = ver or self.latest_ver
        logger.debug("Attempting to download package: %s", self.name)
        # ensure there is a downloadable package release
        download_url = self.download_url(ver)
        if download_url is None:
            return None

        # check if extraction folder exists
        extract_dir = os.path.join(PYPI_SAVE_PATH, self.name + "-" + ver)
        if os.path.isdir(extract_dir):
            if any(os.path.isdir(dirname)
                   for dirname in os.listdir(extract_dir)):
                logger.debug(
                    "Package %s was downloaded already, skipping", self.name)
                return extract_dir  # already extracted
        else:
            os.mkdir(extract_dir)
            self._dirs.append(extract_dir)

        # download file to the folder
        fname = os.path.join(extract_dir, download_url.rsplit("/", 1)[-1])
        try:  # TODO: timeout handling
            urlretrieve(download_url, fname)
        except IOError:  # missing file, very rare but happens
            logger.warning("Broken PyPi link: %s", download_url)
            return None

        # extract using supported format
        extension = ""
        for ext in SUPPORTED_FORMATS:
            if fname.endswith(ext):
                extension = ext
                break
        if not extension:
            raise ValueError("Unexpected archive format: %s" % fname)

        cmd = SUPPORTED_FORMATS[extension] % {
            'fname': fname, 'dir': extract_dir}
        os.system(cmd)

        # fix permissions (+X = traverse dirs)
        os.system('chmod -R u+rwX "%s"' % extract_dir)

        # edge case: zip source archives usually (always?) contain
        # extra level folder. If after extraction there is a single dir in the
        # folder, change extract_dir to that folder
        if download_url.endswith(".zip"):
            single_dir = None
            for entry in os.listdir(extract_dir):
                entry_path = os.path.join(extract_dir, entry)
                if os.path.isdir(entry_path):
                    if single_dir is None:
                        single_dir = entry_path
                    else:
                        single_dir = None
                        break
            if single_dir:
                extract_dir = single_dir

        return extract_dir

    def _info_path(self, ver):
        """
        :return: either xxx.dist-info or xxx.egg-info path, or None

        It is used by dependencies parser and to locate top_level.txt
        """
        extract_dir = self.download(ver)
        if not extract_dir:
            return None

        # hyphens are translated into underscores
        # multiple underscores are collapsed into one
        # there is only one package subject to this rule as of 06/2018
        # https://pypi.org/project/Tzara---A-Personal-Assistant/#files
        #
        # Quote: "Comparison of project names is case insensitive and treats
        # arbitrarily-long runs of underscores, hyphens, and/or periods
        # as equal."
        # This rule was defined in old PyPI Packaging User Guide, but it is not
        # in the actual version. It is quoted here:
        # https://github.com/pypa/pipenv/issues/1302
        cname = re.sub("[_-]+", "_", self.canonical_name)
        dist_info_path = "%s-%s.dist-info" % (cname, ver)
        egg_info_path = "%s.egg-info" % cname
        for info_path in (dist_info_path, egg_info_path, "EGG-INFO"):
            path = os.path.join(extract_dir, info_path)
            if os.path.isdir(path):
                logger.debug("Project has info folder: %s", path)
                return path
        logger.debug(
            "Neither dist-info nor egg-info folders found in %s", self.name)

    @d.cached_method
    def get_setup_params(self, extract_dir=None):
        extract_dir = extract_dir or self.download()
        if not os.path.isfile(os.path.join(extract_dir, 'setup.py')):
            return None
        _, output = shell("docker.sh", extract_dir)
        if not output.strip():
            logger.warning("Could not parse setup() params in %s", extract_dir)
            return None
        return json.loads(output)

    @d.cached_method
    def modules(self, ver=None):
        # type: (str) -> list
        """ Return list of modules provided by this package

        :param ver: str version
        :return: list of modules provided by this package

        For .egg and .whl they're stored in top_level.txt in dist-info path
        For source code, they are in setup() parameters:
            packages: - pure Python modules, the best possible case
            py_modules: - in case of single file packages
            ext_modules: - C extensions

        Tests:
            0.0.1[0.0.1] - tar.gz, dir
            0[0.0.0] - whl, single file (non-importable name)
            02exercicio[1.0.0] - tar.gz, no files
            asciaf[1.0.0] - tar.gz, no files
            0805nexter[1.2.0] - zip, single file
            4suite-xml[1.2.0] - tar.bz2, __init__ folder
            a3rt-sdk-py["0.0.3"] - folder not matching canonical name
            abofly["1.4.0"] - single file, using non-canonical name
        """
        ver = ver or self.latest_ver
        logger.debug("Package %s ver %s top folder:", self.name, ver)
        modules = []  # default return

        extract_dir = self.download(ver)
        if not extract_dir:
            return modules

        info_path = self._info_path(ver)

        def unique(*lists):
            # combine multiple iterables into one list with unique values
            return sorted(set().union(*lists))

        tl_fname = info_path and os.path.join(info_path, 'top_level.txt')
        # egg or wheel package
        if tl_fname and os.path.isfile(tl_fname):
            text = open(tl_fname, 'r').read(1024)
            return unique(
                (line.strip() for line in text.split() if line.strip()))

        # source package - check setup() parameters
        params = self.get_setup_params(extract_dir)
        if params is None:
            return modules
        # scripts are not importable and thus ignored here
        # perhaps they should be considered by module_paths
        keys = ('packages', 'py_modules', 'ext_modules', 'namespaces',
                'namespace_packages')
        return unique(*(params[key] for key in keys if key in params))

    @d.cached_method
    def module_paths(self, ver):
        """ Paths to dirs/files containing provided modules. """
        mod_paths = []
        extract_dir = self.download(ver)
        for path in self.modules(ver):
            if not os.path.isdir(os.path.join(extract_dir, path)):
                if os.path.isfile(os.path.join(extract_dir, path + ".py")):
                    path += ".py"
                else:
                    # a C extension (no path) or a malformed package
                    continue
            mod_paths.append(path)
        return mod_paths

    @d.cached_property
    def repository(self):
        """Search for a pattern in package info and package content
        Search places:
        - info home page field
        - full info page
        - package content
        :return url if found, None otherwise

        >>> Package("numpy").url
        'github.com/numpy/numpy'
        """
        # check home page first
        m = scraper.URL_PATTERN.search(
            self.info.get('info', {}).get('home_page', ""))
        if m:
            return m.group(0)

        pattern = scraper.named_url_pattern(self.name)

        m = re.search(pattern, str(self.info))
        if m:
            return m.group(0)

        for path in self.module_paths(self.latest_ver):
            _, output = shell("zgrep.sh", pattern, path, raise_on_status=False)
            output = output.strip()
            if output:
                return output
        return None

    @d.cached_method
    def dependencies(self, ver=None):
        """Extract dependencies from either wheels metadata or setup.py

        >>> 'numpy' in Package("pandas").dependencies()
        True
        """
        ver = ver or self.latest_ver
        default = {}  # default return vlaue
        logger.debug(
            "Getting dependencies for project %s ver %s", self.name, ver)
        extract_dir = self.download(ver)
        if not extract_dir:
            return default

        info_path = self._info_path(ver) or ""
        if info_path.endswith(".dist-info"):
            logger.debug("    .. WHEEL package, parsing from metadata.json")
            fname = os.path.join(info_path, 'metadata.json')
            if os.path.isfile(fname):
                info = json.load(open(fname))
                # only unconditional dependencies are considered
                # http://legacy.python.org/dev/peps/pep-0426/#dependency-specifiers
                deps = []
                for dep in info.get('run_requires', []):
                    if 'extra' not in dep and 'environment' not in dep:
                        deps.extend(dep['requires'])
            else:
                fname = os.path.join(info_path, 'METADATA')
                if not os.path.isfile(fname):
                    return default
                # example record:
                # Requires-Dist: numpy (>=1.9.0)
                # len("Requires-Dist:") == 14
                raw_deps = [line[14:].strip()
                            for line in open(fname)
                            if line.startswith("Requires-Dist:")]

                deps = []
                for raw_dep in raw_deps:
                    chunks = raw_dep.split(None, 1)
                    if len(chunks) == 1:
                        deps.append(raw_dep)
                        continue
                    chunks[1] = chunks[1].strip("()")
                    deps.append(" ".join(chunks))

        elif info_path.endswith(".egg-info"):
            logger.debug("    .. egg package, parsing requires.txt")
            fname = os.path.join(info_path, 'requires.txt')
            if not os.path.isfile(fname):
                return default
            deps = []
            for line in open(fname, 'r'):
                if "[" in line:
                    break
                if line:
                    deps.append(line)
        else:
            logger.debug("    ..generic package, running setup.py in a sandbox")
            params = self.get_setup_params(extract_dir)
            if params is None:
                logger.debug("    .. looks to be a malformed package")
                return default
            deps = params.get('install_requires', [])

        def dep_split(dependency):
            match = re.match(r"[\w_.-]+", dependency)
            if not match:  # invalid dependency
                name = ""
            else:
                name = match.group(0)
            version = dependency[len(name):].strip()
            return name, version

        return dict(dep_split(dep.strip()) for dep in deps if dep.strip())

    @d.cached_method
    def loc_size(self, ver):
        """get size in LOC"""
        return sum(python_loc_size(path) for path in self.module_paths(ver))



#!/usr/bin/env python

import os
import re
from setuptools import setup

# get author, version and license from package files
# since it has some not yet installed dependencies, parsing the file:
package = 'stecosystems'
head = open(os.path.join(package, '__init__.py')).read(2048)
pattern = r"""__%s__\s*=\s*['"]([^'"]*)['"]"""
kwargs = {keyword: re.search(pattern % keyword, head).group(1)
          for keyword in ('version', 'author', 'license')}

requirements = [
    line.strip()
    for line in open('requirements.txt')
    if line.strip() and not line.strip().startswith('#')]

# options reference: https://docs.python.org/2/distutils/
# see also: https://packaging.python.org/tutorials/distributing-packages/
setup(
    name="strudel.ecosystems",
    description="Ecosystem abstractions for PyPI and npm",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[  # full list: https://pypi.org/classifiers/
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering'
    ],
    platforms=["Linux", "Solaris", "Mac OS-X", "Unix", "Windows"],
    python_requires='>2.6, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    packages=[package],
    url='https://github.com/cmustrudel/strudel.ecosystems',
    install_requires=requirements,
    **kwargs
)




from collections import defaultdict
import json
import os

import pandas as pd

import stutils.email_utils as email
import stscraper as scraper
from stutils import decorators as d
from stutils import mapreduce

from .base import *
from . import pypi
from . import npm

fs_cache = d.fs_cache('npm')


def pypi_packages_info():
    """
    :return: a pd.Dataframe with columns:
        - author: author email, str
        - url: a str suitable for scraper.parse_url() or scraper.get_provider()
        - license: unstructured str to be used with common.utils.parse_license()

    """
    names = []  # list of package names
    urls = {}  # urls[pkgname] = github_url
    authors = {}  # authors[pkgname] = author_email
    licenses = {}
    author_projects = defaultdict(list)
    author_orgs = defaultdict(
        lambda: defaultdict(int))  # orgs[author] = {org: num_packages}

    for package_name in pypi.Package.all():
        logger.info("Processing %s", package_name)
        try:
            p = pypi.Package(package_name)
        except pypi.PackageDoesNotExist:
            # some deleted packages aren't removed from the list
            continue
        names.append(package_name)

        if p.repository:
            urls[package_name] = p.repository

        try:
            author_email = email.clean(p.info["info"].get('author_email'))
        except email.InvalidEmail:
            author_email = None

        if author_email:
            author_projects[author_email].append(package_name)

        authors[package_name] = author_email
        licenses[package_name] = p.info['info']['license']

        if p.repository:
            provider, project_url = scraper.parse_url(p.repository)
            if provider == "github.com":
                org, _ = project_url.split("/")
                author_orgs[author_email][org] += 1

    # at this point, we have ~54K repos
    # by guessing github account from author affiliations we can get 8K more
    processed = 1
    total = len(author_projects)
    for author, packages in author_projects.items():
        logger.info("Postprocessing authors (%d out of %d): %s",
                    processed, total, author)
        processed += 1
        # check all orgs of the author, starting from most used ones
        orgs = [org for org, _ in
                sorted(author_orgs[author].items(), key=lambda x: -x[1])]
        if not orgs:
            continue
        for package in packages:
            if package in urls:
                continue
            for org in orgs:
                url = "%s/%s" % (org, package)
                r = requests.get("https://github.com/" + url)
                if r.status_code == 200:
                    urls[package] = url
                    break

    return pd.DataFrame({"url": urls, "author": authors, 'license': licenses},
                        index=names)


def pypi_dependencies():
    """ Get a bunch of information about npm packages
    This will return pd.DataFrame with package name as index and columns:
        - version: version of release, str
        - date: release date, ISO str
        - deps: names of dependencies, comma separated string
        - raw_dependencies: dependencies, JSON dict name: ver
        - raw_test_dependencies
        - raw_build_dependencies
    """
    deps = {}
    fname = fs_cache.get_cache_fname(".deps_and_size.cache")

    if os.path.isfile(fname):
        logger.info("deps_and_size() cache file already exists. "
                    "Existing records will be reused")

        def gen(df):
            d = {}
            for index, row in df.iterrows():
                item = row.to_dict()
                item["name"] = index[0]
                item["version"] = index[1]
                d[tuple(index)] = item
            return d

        deps = gen(pd.read_csv(fname, index_col=["name", "version"]))

    else:
        logger.info("deps_and_size() cache file doesn't exists. "
                    "Computing everything from scratch is a lengthy process "
                    "and will likely take a week or so")

    tp = mapreduce.ThreadPool()
    logger.info("Starting a threadppol with %d workers...", tp.n)

    package_names = pypi_packages_info().index

    def do(pkg_name, ver, release_date):
        # this method is used by worker threads, calling done() on finish
        p_deps = pypi.Package(pkg_name).dependencies(ver)

        return {
            'name': pkg_name,
            'version': ver,
            'date': release_date,
            'deps': ",".join(p_deps.keys()).lower(),
            'raw_dependencies': json.dumps(p_deps)
        }

    def done(output):
        deps[(output["name"], output["version"])] = output

    for package_name in package_names:
        logger.info("Processing %s", package_name)
        try:
            p = pypi.Package(package_name)
        except pypi.PackageDoesNotExist:
            continue

        for version, release_date in p.releases(True, True):
            if (package_name, version) not in deps:
                logger.info("    %s", version)
                tp.submit(do, package_name, version, release_date, callback=done)
            else:
                logger.info("    %s (cached)", version)

    # wait for workers to complete
    tp.shutdown()

    # save updates
    df = pd.DataFrame(deps.values()).sort_values(["name", "version"]).set_index(
        ["name", "version"], drop=True)
    df.to_csv(fname)

    return df


@fs_cache
def npm_packages_info():
    # type: () -> pd.DataFrame
    """ Get a bunch of information about npm packages
    This will return pd.DataFrame with package name as index and columns:
        - url: date of release, YYYY-MM-DD str
        - version: version of release, str
        - deps: dependencies, comma separated string
        - owners
    """

    def gen():
        logger = logging.getLogger("npm.utils.package_info")
        for package in npm.Package.all():
            logger.info("Processing %s", package['key'])
            # TODO: before falling back to str(package), use named pattern
            repo = npm._get_field(package['doc'].get('repository'), 'url') or \
                npm._get_field(package['doc'].get('homepage'), 'url') or \
                npm._get_field(package['doc'].get('bugs'), 'url') or str(package)

            m = repo and scraper.URL_PATTERN.search(repo)

            yield {
                'name': package['key'],
                'url': m and m.group(0),
                'author': npm._get_field(package['doc'].get('author', {}), 'email'),
                'license': npm.json_path(package, 'doc', 'license')
            }

    return pd.DataFrame(gen()).set_index('name', drop=True)


def npm_dependencies():
    """ Get a bunch of information about npm packages
    This will return pd.DataFrame with package name as index and columns:
        - version: version of release, str
        - date: release date, ISO str
        - deps: names of dependencies, comma separated string
        - raw_dependencies: dependencies, JSON dict name: ver
    """

    def gen():
        logger = logging.getLogger("npm.utils.package_info")
        for package in npm.Package.all():
            logger.info("Processing %s", package['key'])
            # possible sources of release date:
            # - ['doc']['time'][<ver>] - best source, sometimes missing
            # - ['doc']['versions'][<ver>]['ctime|mtime']  # e.g. Graph
            # - ['doc']['time']['modified|created'] # e.g. stack-component
            # - ['doc']['ctime|mtime']  # e.g. Lingo
            # - empty  # JSLint-commonJS

            for version, release in package['doc'].get('versions', {}).items():
                deps = release.get('dependencies') or {}
                deps = {dep.decode("utf8"): ver
                        for dep, ver in deps.items()}
                time = npm.json_path(package, 'doc', 'time', version) or \
                    npm.json_path(release, 'ctime') or \
                    npm.json_path(release, 'mtime') or \
                    npm.json_path(package, 'doc', 'time', 'created') or \
                    npm.json_path(package, 'doc', 'time', 'modified') or \
                    None

                yield {
                    'name': package['key'],
                    'version': version,
                    'date': time,
                    'deps': ",".join(deps.keys()),
                    'raw_dependencies': json.dumps(deps)
                }

    return pd.DataFrame(gen()).sort_values(
        ['name', 'date']).set_index('name', drop=True)
